#include "inputCommon.h"
#include "IInputInfo.h"
#include <windows.h>
#include <vfw.h>
#include <msacm.h>
#include <cstdlib>
#include <cstring>
#include <string>

using namespace System;
using namespace System::IO;
using namespace System::Drawing;
using namespace System::Drawing::Imaging;
using namespace System::Drawing::Drawing2D;
using namespace System::Collections::Generic;
using namespace System::Runtime::InteropServices;
using namespace System::Windows::Forms;

namespace ruche {
namespace aviutl {
namespace plugins
{
    // AviUtl̓̓vOCbvNXB
    public ref class InputPlugin
    {
      private:
        // fBAf[^̓͏񋟂NXB
        ref class InputInfo : IInputInfo
        {
          private:
            InputMediaFlags _types;
            int _imageWidth;
            int _imageHeight;
            int _imageFrameRateBase;
            int _imageFrameRateScale;
            int _imageFrameCount;
            int _audioSampleCount;

          public:
            // RXgN^B
            // info : ͏B
            InputInfo(const INPUT_INFO& info)
            {
                _types = static_cast<InputMediaFlags>(info.flag);
                _imageWidth = (info.format == NULL) ?
                    0 : Math::Abs(info.format->biWidth);
                _imageHeight = (info.format == NULL) ?
                    0 : Math::Abs(info.format->biHeight);
                _imageFrameRateBase = info.rate;
                _imageFrameRateScale = info.scale;
                _imageFrameCount = info.n;
                _audioSampleCount = 0;
                if (info.audio_format != NULL)
                {
                    DWORD srcRate = info.audio_format->nSamplesPerSec;
                    long long srcCount = static_cast<long long>(info.audio_n);
                    _audioSampleCount =
                        static_cast<int>((srcCount * 48000LL) / srcRate);
                }
            }

            // ͂f[^^CṽtOl擾B
            virtual property InputMediaFlags InputTypes
            {
                InputMediaFlags get() { return _types; }
            }

            // ͂摜̉擾B
            virtual property int InputImageWidth
            {
                int get() { return _imageWidth; }
            }

            // ͂摜̏c擾B
            virtual property int InputImageHeight
            {
                int get() { return _imageHeight; }
            }

            // ͂摜̃t[[g̕ql擾B
            virtual property int InputImageFrameRateBase
            {
                int get() { return _imageFrameRateBase; }
            }

            // ͂摜̃t[[g̕l擾B
            virtual property int InputImageFrameRateScale
            {
                int get() { return _imageFrameRateScale; }
            }

            // ͂摜̑t[擾B
            virtual property int InputImageFrameCount
            {
                int get() { return _imageFrameCount; }
            }

            // ͂鉹̑TvO擾B
            // 48kHz, 16rbg, XeIɕϊ̒lł邱ƂɒӁB
            virtual property int InputAudioSampleCount
            {
                int get() { return _audioSampleCount; }
            }
        };

        // rfIR[fbN\񋓃tOlB
        enum class VideoCodec
        {
            // 񈳏k
            Rgb = 0,

            // YUY2
            Yuy2,

            // PIXEL_YC
            PixelYc,

            // ̑̃R[fbN
            Others,
        };

      private:
        // W[̃tpXB
        String^ _filePath;

        // W[̃t@CB
        String^ _fileName;

        // W[nhB
        HMODULE _hPlugin;

        // vOCC^tF[XB
        INPUT_PLUGIN_TABLE* _pi;

        // vOCς݃tOB
        bool _pluginInitialized;

        // ̓T|[g^CvB
        InputFlags _types;

        // vOCB
        String^ _pluginName;

        // t@CtB^B
        String^ _fileFilter;

        // vOCB
        String^ _pluginInfo;

        // YUY2̃XP[O^CvB
        Yuy2ScaleType _yuy2Scale;

        // JĂ郁fBAt@C̃nhB
        INPUT_HANDLE _mediaHandle;

        // JĂ郁fBAt@C̃tpXB
        String^ _mediaFilePath;

        // JĂ郁fBAt@C̓͏B
        INPUT_INFO* _mediaInfoBase;
        InputInfo^ _mediaInfo;

        // JĂ郁fBAt@C̉摜R[fbNB
        VideoCodec _mediaImageCodec;

        // JĂ郁fBAt@C̉摜R[fbNnhB
        HIC _mediaImageCodecHandle;

        // JĂ郁fBAt@C̓WJp摜tH[}bgB
        BITMAPV4HEADER* _mediaImageDecompressFormat;

        // JĂ郁fBAt@C̉摜t[擾pobt@B
        array<Byte>^ _mediaImageBuffer;

        // JĂ郁fBAt@C̓WJ摜i[pobt@B
        array<Byte>^ _mediaImageDecompressBuffer;

        // JĂ郁fBAt@C̍sXbvpobt@B
        array<Byte>^ _mediaImageLineSwapBuffer;

        // JĂ郁fBAt@C̉Xg[nhB
        HACMSTREAM _mediaAudioHandle;

        // JĂ郁fBAt@C̓WJptH[}bgB
        WAVEFORMATEX* _mediaAudioDecompressFormat;

        // JĂ郁fBAt@C̉Tv擾pobt@B
        array<Byte>^ _mediaAudioBuffer;

        // JĂ郁fBAt@C̓WJi[pobt@B
        array<Byte>^ _mediaAudioDecompressBuffer;

      public:
        // RXgN^B
        // auiPath : ̓vOC̃t@CpXB
        InputPlugin(String^ auiPath)
        {
            // ȍ
            _hPlugin = NULL;
            _pi = NULL;
            _pluginInitialized = false;
            _yuy2Scale = Yuy2ScaleType::Full;
            _mediaHandle = NULL;
            _mediaFilePath = nullptr;
            _mediaInfoBase = NULL;
            _mediaInfo = nullptr;
            _mediaImageCodec = VideoCodec::Rgb;
            _mediaImageCodecHandle = NULL;
            _mediaImageDecompressFormat = NULL;
            _mediaImageBuffer = nullptr;
            _mediaImageDecompressBuffer = nullptr;
            _mediaImageLineSwapBuffer = nullptr;
            _mediaAudioHandle = NULL;
            _mediaAudioDecompressFormat = NULL;
            _mediaAudioBuffer = nullptr;
            _mediaAudioDecompressBuffer = nullptr;

            try
            {
                if (auiPath == nullptr)
                {
                    throw gcnew ArgumentNullException("auiPath");
                }

                // std::wstring ɕϊ
                pin_ptr<wchar_t> pathPin = &auiPath->ToCharArray()[0];
                std::wstring path(pathPin, auiPath->Length);
                pathPin = nullptr;

                // W[擾
                _hPlugin = ::LoadLibraryW(path.data());
                if (_hPlugin == NULL)
                {
                    throw gcnew ArgumentException(
                        auiPath + " ͗LAviUtl̓vOCł͂܂B",
                        "auiPath");
                }

                // API֐擾
                Func_GetInputPluginTable getInputPluginTable =
                    reinterpret_cast<Func_GetInputPluginTable>(
                        ::GetProcAddress(_hPlugin, "GetInputPluginTable"));
                if (getInputPluginTable == NULL)
                {
                    throw gcnew ArgumentException(
                        auiPath + " ͗LAviUtl̓vOCł͂܂B",
                        "auiPath");
                }

                // vOCC^tF[X擾
                _pi = getInputPluginTable();
                if (
                    _pi                == NULL ||
                    _pi->func_open     == NULL ||
                    _pi->func_close    == NULL ||
                    _pi->func_info_get == NULL)
                {
                    throw gcnew ApplicationException(
                        auiPath + " C^tF[X擾ł܂łB");
                }

                // vOC
                if (_pi->func_init != NULL && _pi->func_init() == FALSE)
                {
                    throw gcnew ApplicationException(
                        auiPath + " ̏Ɏs܂B");
                }
                _pluginInitialized = true;

                // tpXt@Cݒ
                _filePath = auiPath;
                _fileName = Path::GetFileName(auiPath);

                // o̓^Cv擾
                _types = static_cast<InputFlags>(_pi->flag);

                // vOCݒ
                _pluginName = gcnew String(_pi->name);
                _pluginInfo = gcnew String(_pi->information);

                // tB^ݒ(\0 ؂肩 | ؂ɕύX)
                std::string filter;
                for (const char* s = _pi->filefilter; *s != '\0'; )
                {
                    filter += s;
                    s += ::strlen(s) + 1;
                    if (*s != '\0') { filter += "|"; }
                }
                _fileFilter = gcnew String(filter.data());

                // 쐬
                _mediaInfoBase = new INPUT_INFO;
                _mediaImageDecompressFormat = new BITMAPV4HEADER;
                _mediaAudioDecompressFormat = new WAVEFORMATEX;
            }
            catch (...)
            {
                this->!InputPlugin();

                throw;
            }
        }

        // t@CiCUB
        !InputPlugin()
        {
            // fBAt@C
            CloseMediaFile();

            // 
            if (_mediaAudioDecompressFormat != NULL)
            {
                delete _mediaAudioDecompressFormat;
                _mediaAudioDecompressFormat = NULL;
            }
            if (_mediaImageDecompressFormat != NULL)
            {
                delete _mediaImageDecompressFormat;
                _mediaImageDecompressFormat = NULL;
            }
            if (_mediaInfoBase != NULL)
            {
                delete _mediaInfoBase;
                _mediaInfoBase = NULL;
            }

            // vOC
            if (_pi != NULL)
            {
                if (_pluginInitialized && _pi->func_exit != NULL)
                {
                    _pi->func_exit();
                }
                _pluginInitialized = false;
                _pi = NULL;
            }

            // W[nh
            if (_hPlugin != NULL)
            {
                ::FreeLibrary(_hPlugin);
                _hPlugin = NULL;
            }
        }

        // fXgN^(Dispose)B
        ~InputPlugin()
        {
            this->!InputPlugin();
        }

      private:
        // W̉摜tH[}bg쐬B
        // width, height : 摜̏cB
        // topDown : gbv_E`ɂȂ true B
        // format : 쐬tH[}bg̐ݒB
        static void MakeStandardImageFormat(
            long width, long height, bool topDown, BITMAPV4HEADER* format)
        {
            // `FbN
            if (format == NULL)
            {
                throw gcnew ArgumentNullException("format");
            }
            if (width <= 0)
            {
                throw gcnew ArgumentOutOfRangeException(
                    "width", width, "width  0 ȉłB");
            }
            if (height <= 0)
            {
                throw gcnew ArgumentOutOfRangeException(
                    "height", height, "height  0 ȉłB");
            }

            // tH[}bg쐬(32rbgARGB)
            ::memset(format, 0, sizeof(BITMAPV4HEADER));
            format->bV4Size = sizeof(BITMAPV4HEADER);
            format->bV4Width = width;
            format->bV4Height = topDown ? -height : height;
            format->bV4Planes = 1;
            format->bV4BitCount = 32;
            format->bV4V4Compression = BI_BITFIELDS;
            format->bV4SizeImage = (width * 4) * height;
            format->bV4RedMask   = 0x00FF0000UL;
            format->bV4GreenMask = 0x0000FF00UL;
            format->bV4BlueMask  = 0x000000FFUL;
            format->bV4AlphaMask = 0xFF000000UL;
            format->bV4CSType = LCS_sRGB;
        }

        // 摜tH[}bggbv_E`ł邩ۂ擾B
        // format : 摜tH[}bgB
        static bool IsTopDownImageFormat(const BITMAPINFOHEADER* format)
        {
            // `FbN
            if (format == NULL)
            {
                throw gcnew ArgumentNullException("format");
            }

            return (format->biHeight < 0);
        }

        // 摜tH[}bggbv_E`ł邩ۂ擾B
        // format : 摜tH[}bgB
        static bool IsTopDownImageFormat(const BITMAPV4HEADER* format)
        {
            // `FbN
            if (format == NULL)
            {
                throw gcnew ArgumentNullException("format");
            }

            return (format->bV4Height < 0);
        }

        // R[fbNnhIDrfIR[fbNl擾B
        // handler : R[fbNnhIDB
        static VideoCodec GetVideoCodec(DWORD handler)
        {
            VideoCodec codec = VideoCodec::Others;

            switch (handler)
            {
              case BI_RGB:
                codec = VideoCodec::Rgb;
                break;

              case mmioFOURCC('Y', 'U', 'Y', '2'):
                codec = VideoCodec::Yuy2;
                break;

              case mmioFOURCC('Y', 'C', '4', '8'):
                codec = VideoCodec::PixelYc;
                break;

              default:
                codec = VideoCodec::Others;
                break;
            }

            return codec;
        }

        // wb_ƍvTCYsNZtH[}bgZoB
        // sȏꍇ PixelFormat::Undefined ԂB
        // header : wb_B
        // infoSize : vTCYB
        static PixelFormat CalcPixelFormat(
            const BITMAPINFOHEADER* header, DWORD infoSize)
        {
            PixelFormat format = PixelFormat::Undefined;

            // rbgtB[htwb_̍ŒTCY
            DWORD bfHeadSize = sizeof(BITMAPINFOHEADER) + sizeof(DWORD) * 3;
            DWORD bfAlphaHeadSize = bfHeadSize + sizeof(DWORD);

            // sNZtH[}bgI
            switch (header->biBitCount)
            {
              case 1:
                if (infoSize >= header->biSize + sizeof(RGBQUAD) * 2)
                {
                    format = PixelFormat::Format1bppIndexed;
                }
                break;

              case 4:
                if (infoSize >= header->biSize + sizeof(RGBQUAD) * 16)
                {
                    format = PixelFormat::Format4bppIndexed;
                }
                break;

              case 8:
                if (infoSize >= header->biSize + sizeof(RGBQUAD) * 256)
                {
                    format = PixelFormat::Format8bppIndexed;
                }
                break;

              case 16:
                switch (header->biCompression)
                {
                  case BI_RGB:
                    format = PixelFormat::Format16bppRgb555;
                    break;

                  case BI_BITFIELDS:
                    if (infoSize >= bfHeadSize)
                    {
                        // rbgtB[h擾
                        const unsigned char* headPtr =
                            reinterpret_cast<const unsigned char*>(header);
                        const DWORD* bf = reinterpret_cast<const DWORD*>(
                            headPtr + header->biSize);

                        if (
                            bf[0] == 0x00007C00 &&
                            bf[1] == 0x000003E0 &&
                            bf[2] == 0x0000001F)
                        {
                            if (
                                header->biSize >= bfAlphaHeadSize &&
                                bf[3] == 0x00008000)
                            {
                                format = PixelFormat::Format16bppArgb1555;
                            }
                            else
                            {
                                format = PixelFormat::Format16bppRgb555;
                            }
                        }
                        else if (
                            bf[0] == 0x0000F800 &&
                            bf[1] == 0x000007E0 &&
                            bf[2] == 0x0000001F)
                        {
                            format = PixelFormat::Format16bppRgb565;
                        }
                    }
                    break;
                }
                break;

              case 24:
              case 0:
                format = PixelFormat::Format24bppRgb;
                break;

              case 32:
                switch (header->biCompression)
                {
                  case BI_RGB:
                    format = PixelFormat::Format32bppRgb;
                    break;

                  case BI_BITFIELDS:
                    if (infoSize >= bfHeadSize)
                    {
                        // rbgtB[h擾
                        const unsigned char* headPtr =
                            reinterpret_cast<const unsigned char*>(header);
                        const DWORD* bf = reinterpret_cast<const DWORD*>(
                            headPtr + header->biSize);

                        if (
                            bf[0] == 0x00FF0000 &&
                            bf[1] == 0x0000FF00 &&
                            bf[2] == 0x000000FF)
                        {
                            if (
                                header->biSize >= bfAlphaHeadSize &&
                                bf[3] == 0xFF000000)
                            {
                                format = PixelFormat::Format32bppArgb;
                            }
                            else
                            {
                                format = PixelFormat::Format32bppRgb;
                            }
                        }
                    }
                    break;
                }
                break;
            }

            return format;
        }

      private:
        // OpenMediaFile ĂяoAfp̏sB
        bool OpenMediaFileVideo()
        {
            // Kvȃp[^Ύs
            if (
                _mediaInfoBase->format == NULL ||
                _pi->func_read_video == NULL)
            {
                return false;
            }

            // 擾
            long width  = Math::Abs(_mediaInfoBase->format->biWidth);
            long height = Math::Abs(_mediaInfoBase->format->biHeight);

            // 摜R[fbN擾
            DWORD codecId = _mediaInfoBase->handler;
            _mediaImageCodec = GetVideoCodec(codecId);
            if (_mediaImageCodec == VideoCodec::Rgb)
            {
                // biCompression ɃnhIDĂꍇ
                switch (_mediaInfoBase->format->biCompression)
                {
                  case BI_RGB:
                  case BI_RLE8:
                  case BI_RLE4:
                  case BI_BITFIELDS:
                  case BI_JPEG:
                  case BI_PNG:
                    // ʏDIBȂ̂őfʂ
                    break;

                  default:
                    codecId = _mediaInfoBase->format->biCompression;
                    _mediaImageCodec = GetVideoCodec(codecId);
                    break;
                }
            }

            // WJp摜tH[}bg쐬(gbv_E`)
            MakeStandardImageFormat(
                width, height, true, _mediaImageDecompressFormat);

            // R[fbN Others ̏ꍇAR[fbNnh쐬
            if (_mediaImageCodec == VideoCodec::Others)
            {
                BITMAPINFOHEADER* outHead =
                    reinterpret_cast<BITMAPINFOHEADER*>(
                        _mediaImageDecompressFormat);
                _mediaImageCodecHandle = ::ICLocate(
                    ICTYPE_VIDEO,
                    codecId,
                    _mediaInfoBase->format,
                    outHead,
                    ICMODE_DECOMPRESS);
                if (_mediaImageCodecHandle == NULL)
                {
                    // s{gAbv`ɂčēx
                    MakeStandardImageFormat(
                        width, height, false, _mediaImageDecompressFormat);
                    _mediaImageCodecHandle = ::ICLocate(
                        ICTYPE_VIDEO,
                        codecId,
                        _mediaInfoBase->format,
                        outHead,
                        ICMODE_DECOMPRESS);
                    if (_mediaImageCodecHandle == NULL)
                    {
                        return false;
                    }

                    // {gAbv`Ȃ΃gbv_E`ւ̕ϊKvɂȂ邽߁A
                    // sXbvpobt@pӂĂ
                    WORD bitCount = _mediaImageDecompressFormat->bV4BitCount;
                    _mediaImageLineSwapBuffer = gcnew array<Byte>(
                        (((width * bitCount + 31) / 32) * 4) * height);
                }
            }

            // 摜t[擾pobt@쐬
#if 0
            DWORD bufSize = _mediaInfoBase->format->biSizeImage;
            if (bufSize == 0)
            {
                // TCYw肳ĂȂꍇ͎OŌvZ
                WORD bitCount = _mediaInfoBase->format->biBitCount;
                if (bitCount == 0)
                {
                    bitCount =
                        (_mediaImageCodec == VideoCodec::PixelYc) ? 48 : 24;
                }
                bufSize = (((width * bitCount + 31) / 32) * 4) * height;
            }
#else
            // mp4input.auiŕsANZXɂȂ邱Ƃ̂
            // I48rbgvZŃobt@쐬Ă݂B
            WORD bitCount = _mediaInfoBase->format->biBitCount;
            if (bitCount < 48) { bitCount = 48; }
            DWORD bufSize = (((width * bitCount + 31) / 32) * 4) * height;
#endif
            _mediaImageBuffer = gcnew array<Byte>(bufSize);

            // 摜t[WJobt@쐬
            _mediaImageDecompressBuffer =
                gcnew array<Byte>(_mediaImageDecompressFormat->bV4SizeImage);

            return true;
        }

        // CloseMediaFile ĂяoAfp̉sB
        void CloseMediaFileVideo()
        {
            if (_mediaImageDecompressBuffer != nullptr)
            {
                _mediaImageDecompressBuffer = nullptr;
            }
            if (_mediaImageBuffer != nullptr)
            {
                _mediaImageBuffer = nullptr;
            }
            if (_mediaImageLineSwapBuffer != nullptr)
            {
                _mediaImageLineSwapBuffer = nullptr;
            }
            if (_mediaImageCodecHandle != NULL)
            {
                ::ICClose(_mediaImageCodecHandle);
                _mediaImageCodecHandle = NULL;
            }
        }

        // OpenMediaFile ĂяoAp̏sB
        bool OpenMediaFileAudio()
        {
            // Kvȃp[^Ύs
            if (
                _mediaInfoBase->audio_format == NULL ||
                _pi->func_read_audio == NULL)
            {
                return false;
            }

            // WJptH[}bg쐬
            WAVEFORMATEX* format = _mediaAudioDecompressFormat;
            format->wFormatTag = WAVE_FORMAT_PCM;
            format->nChannels = 2;
            format->nSamplesPerSec = 48000;
            format->wBitsPerSample = 16;
            format->nBlockAlign =
                (format->wBitsPerSample / 8) * format->nChannels;
            format->nAvgBytesPerSec =
                format->nSamplesPerSec * format->nBlockAlign;
            format->cbSize = 0;

            // Xg[nh쐬
            pin_ptr<HACMSTREAM> handlePtr = &_mediaAudioHandle;
            MMRESULT res = ::acmStreamOpen(
                handlePtr, NULL,
                _mediaInfoBase->audio_format, format,
                NULL, 0, 0, ACM_STREAMOPENF_NONREALTIME);
            handlePtr = nullptr;
            if (res != 0)
            {
                _mediaAudioHandle = NULL;
                return false;
            }

            return true;
        }

        // CloseMediaFile ĂяoAp̉sB
        void CloseMediaFileAudio()
        {
            if (_mediaAudioDecompressBuffer != nullptr)
            {
                _mediaAudioDecompressBuffer = nullptr;
            }
            if (_mediaAudioBuffer != nullptr)
            {
                _mediaAudioBuffer = nullptr;
            }
            if (_mediaAudioHandle != NULL)
            {
                ::acmStreamClose(_mediaAudioHandle, 0);
                _mediaAudioHandle = NULL;
            }
        }

        // 񈳏k摜f[^WJB
        bool DecompressRgbImage()
        {
            bool result = false;

            Bitmap^ bmpOrg = nullptr;
            Bitmap^ bmp32 = nullptr;
            Graphics^ g = nullptr;

            try
            {
                const BITMAPINFOHEADER* head = _mediaInfoBase->format;

                // sNZtH[}bg𓾂
                PixelFormat format =
                    CalcPixelFormat(head, _mediaInfoBase->format_size);
                if (format == PixelFormat::Undefined)
                {
                    return false;
                }

                // rbg𓾂
                DWORD bitCount = head->biBitCount;
                if (bitCount == 0) { bitCount = 24; }

                // rbg}bv쐬
                long width = Math::Abs(_mediaImageDecompressFormat->bV4Width);
                long height = Math::Abs(_mediaImageDecompressFormat->bV4Height);
                long stride = ((width * bitCount + 31) / 32) * 4;
                pin_ptr<unsigned char> bufPtr = &_mediaImageBuffer[0];
                IntPtr bufIntPtr(bufPtr);
                bmpOrg = gcnew Bitmap(width, height, stride, format, bufIntPtr);
                bufPtr = nullptr;

                // pbgݒ
                if (head->biBitCount <= 8)
                {
                    // pbgAhX擾
                    const unsigned char* headPtr =
                        reinterpret_cast<const unsigned char*>(head);
                    const RGBQUAD* palSrc =
                        reinterpret_cast<const RGBQUAD*>(headPtr + head->biSize);
                    int palCount = 1 << bitCount;

                    // ݒpbg擾
                    ColorPalette^ palDest = bmpOrg->Palette;
                    palCount = min(palDest->Entries->Length, palCount);

                    // ݒ
                    for (int i = 0; i < palCount; ++i)
                    {
                        palDest->Entries[i] = Color::FromArgb(
                            palSrc[i].rgbRed,
                            palSrc[i].rgbGreen,
                            palSrc[i].rgbBlue);
                    }
                    bmpOrg->Palette = palDest;
                }

                // 32rbgARGB`ɕϊ
                if (bmpOrg->PixelFormat == PixelFormat::Format32bppArgb)
                {
                    bmp32 = bmpOrg;
                }
                else
                {
                    // 傫32rbgARGBrbg}bv쐬
                    bmp32 = gcnew Bitmap(
                        bmpOrg->Width,
                        bmpOrg->Height,
                        PixelFormat::Format32bppArgb);

                    // OtBNXIuWFNg擾
                    g = Graphics::FromImage(bmp32);

                    // 摜`
                    g->CompositingMode = CompositingMode::SourceCopy;
                    g->DrawImage(bmpOrg, 0, 0);

                    // OtBNXIuWFNg
                    delete g;
                    g = nullptr;
                }

                // 摜{gAbv`ȂΏ㉺]
                if (!IsTopDownImageFormat(head))
                {
                    bmp32->RotateFlip(RotateFlipType::RotateNoneFlipY);
                }

                // z֏o
                BitmapData^ srcData = bmp32->LockBits(
                    System::Drawing::Rectangle(
                        0, 0, bmp32->Width, bmp32->Height),
                    ImageLockMode::ReadOnly,
                    PixelFormat::Format32bppArgb);
                Marshal::Copy(
                    srcData->Scan0, _mediaImageDecompressBuffer,
                    0, _mediaImageDecompressFormat->bV4SizeImage);
                bmp32->UnlockBits(srcData);

                result = true;
            }
            catch (...)
            {
                result = false;
            }
            finally
            {
                // \[X̔j
                if (g != nullptr)
                {
                    delete g;
                    g = nullptr;
                }
                if (bmp32 != nullptr)
                {
                    if (bmp32 != bmpOrg) { delete bmp32; }
                    bmp32 = nullptr;
                }
                if (bmpOrg != nullptr)
                {
                    delete bmpOrg;
                    bmpOrg = nullptr;
                }
            }

            return result;
        }

        // YUY2`摜f[^WJB
        bool DecompressYuy2Image()
        {
            int sy[2], su, sv, yy, uu, vv, r, g, b;

            // ݒ
            long width = Math::Abs(_mediaImageDecompressFormat->bV4Width);
            long height = Math::Abs(_mediaImageDecompressFormat->bV4Height);
            int bitCount = _mediaImageDecompressFormat->bV4BitCount;
            if (bitCount == 0) { bitCount = 24; }
            long destLineLen = ((width * bitCount + 31) / 32) * 4;
            long srcLineLen = ((width + 1) / 2) * 4;

            // o͐|C^擾
            pin_ptr<unsigned char> srcPtr = &_mediaImageBuffer[0];
            pin_ptr<unsigned char> destPtr = &_mediaImageDecompressBuffer[0];

            // sPʂŏ
            for (long y = 0; y < height; ++y)
            {
                // ++ Ouł悤1ÕAhX擾
                const unsigned char* ps = &srcPtr[y * srcLineLen] - 1;
                unsigned char* pd = &destPtr[y * destLineLen] - 1;

                // 2sNZ
                for (int x = 0; x < width; x += 2)
                {
                    // YUY2p[^擾
                    sy[0] = *++ps;
                    su    = *++ps;
                    sy[1] = *++ps;
                    sv    = *++ps;

                    int loopCount = (x >= width - 1) ? 1 : 2;
                    for (int i = 0; i < loopCount; ++i)
                    {
                        // RGBlvZ
                        switch (Yuy2Scale)
                        {
                          case Yuy2ScaleType::Compressed:
                            // R = 1.164*(Y-16)                 + 1.596*(V-128)
                            // G = 1.164*(Y-16) - 0.391*(U-128) - 0.813*(V-128)
                            // B = 1.164*(Y-16) + 2.018*(U-128)
                            yy = sy[i] - 16;
                            uu = su - 128;
                            vv = sv - 128;
                            r = (1164 * yy             + 1596 * vv + 500) / 1000;
                            g = (1164 * yy -  391 * uu -  813 * vv + 500) / 1000;
                            b = (1164 * yy + 2018 * uu             + 500) / 1000;
                            break;

                          case Yuy2ScaleType::Full:
                          default:
                            // R = 1.000*Y                   + 1.402*(V-127.5)
                            // G = 1.000*Y - 0.344*(U-127.5) - 0.714*(V-127.5)
                            // B = 1.000*Y + 1.772*(U-127.5)
                            uu = su * 2 - 255;
                            vv = sv * 2 - 255;
                            r = (1000 * sy[i]            + 701 * vv + 500) / 1000;
                            g = (1000 * sy[i] - 172 * uu - 357 * vv + 500) / 1000;
                            b = (1000 * sy[i] + 886 * uu            + 500) / 1000;
                            break;
                        }

                        // lݒ
                        *++pd = static_cast<unsigned char>(min(max(b, 0), 255));
                        *++pd = static_cast<unsigned char>(min(max(g, 0), 255));
                        *++pd = static_cast<unsigned char>(min(max(r, 0), 255));
                        *++pd = 255; // At@l
                    }
                }
            }

            destPtr = nullptr;
            srcPtr = nullptr;

            return true;
        }

        // PIXEL_YC`摜f[^WJB
        bool DecompressPixelYcImage()
        {
            int sy, su, sv, r, g, b;

            // ݒ
            long width = Math::Abs(_mediaImageDecompressFormat->bV4Width);
            long height = Math::Abs(_mediaImageDecompressFormat->bV4Height);
            int bitCount = _mediaImageDecompressFormat->bV4BitCount;
            if (bitCount == 0) { bitCount = 24; }
            long destLineLen = ((width * bitCount + 31) / 32) * 4;
            long srcLineLen = width * 6;

            // o͐|C^擾
            pin_ptr<unsigned char> srcPtr = &_mediaImageBuffer[0];
            pin_ptr<unsigned char> destPtr = &_mediaImageDecompressBuffer[0];

            // sPʂŏ
            for (long y = 0; y < height; ++y)
            {
                // ++ Ouł悤1ÕAhX擾
                const short* ps =
                    reinterpret_cast<const short*>(&srcPtr[y * srcLineLen]) - 1;
                unsigned char* pd = &destPtr[y * destLineLen] - 1;

                // 1sNZ
                for (int x = 0; x < width; ++x)
                {
                    // PIXEL_YCp[^擾
                    sy = *++ps;
                    su = *++ps;
                    sv = *++ps;

                    // RGBlvZ
#if 0
                    // R = 0.0623*Y            + 0.0873*V
                    // G = 0.0623*Y - 0.0214*U - 0.0445*V
                    // B = 0.0623*Y + 0.1103*U
                    r = (623 * sy             + 873 * sv + 5000) / 10000;
                    g = (623 * sy -  214 * su - 445 * sv + 5000) / 10000;
                    b = (623 * sy + 1103 * su            + 5000) / 10000;
#else
                    // http://mksoft.hp.infoseek.co.jp/doc/aviutlyc.html 𗘗p
                    r = (255 * sy + ((                     ( 22881 * sv >> 16) + 3) << 10)) >> 12;
                    g = (255 * sy + (((-5616 * su >> 16) + (-11655 * sv >> 16) + 3) << 10)) >> 12;
                    b = (255 * sy + (((28919 * su >> 16)                       + 3) << 10)) >> 12;
#endif

                    // lݒ(0`255Ɋۂ߂)
                    *++pd = static_cast<unsigned char>(min(max(b, 0), 255));
                    *++pd = static_cast<unsigned char>(min(max(g, 0), 255));
                    *++pd = static_cast<unsigned char>(min(max(r, 0), 255));
                    *++pd = 255; // At@l
                }
            }

            destPtr = nullptr;
            srcPtr = nullptr;

            return true;
        }

        // ̑̃R[fbN`摜f[^WJB
        bool DecompressOthersImage()
        {
            BITMAPINFOHEADER* outHead =
                reinterpret_cast<BITMAPINFOHEADER*>(_mediaImageDecompressFormat);

            // WJ
            pin_ptr<unsigned char> srcPtr = &_mediaImageBuffer[0];
            pin_ptr<unsigned char> destPtr = &_mediaImageDecompressBuffer[0];
            DWORD resDecomp = ::ICDecompress(
                _mediaImageCodecHandle,
                ICDECOMPRESS_HURRYUP,
                _mediaInfoBase->format,
                srcPtr,
                outHead,
                destPtr);
            destPtr = nullptr;
            srcPtr = nullptr;
            bool result = (resDecomp == ICERR_OK);

            // {gAbv`ȂΏ㉺]
            if (result && !IsTopDownImageFormat(_mediaImageDecompressFormat))
            {
                long width = Math::Abs(_mediaImageDecompressFormat->bV4Width);
                long height = Math::Abs(_mediaImageDecompressFormat->bV4Height);
                int bitCount = _mediaImageDecompressFormat->bV4BitCount;
                if (bitCount == 0) { bitCount = 24; }
                long lineSize = ((width * bitCount + 31) / 32) * 4;

                long h_2 = height / 2;
                for (long y = 0; y < h_2; ++y)
                {
                    long sy = height - y - 1;
                    Array::Copy(
                        _mediaImageDecompressBuffer, lineSize * y,
                        _mediaImageLineSwapBuffer, 0, lineSize);
                    Array::Copy(
                        _mediaImageDecompressBuffer, lineSize * sy,
                        _mediaImageDecompressBuffer, lineSize * y, lineSize);
                    Array::Copy(
                        _mediaImageLineSwapBuffer, 0,
                        _mediaImageDecompressBuffer, lineSize * sy, lineSize);
                }
            }

            return true;
        }

        // ̏o̓Tvlf[^̃TvlZoB
        int CalcSourceAudioSampleValue(int destSampleValue)
        {
            return static_cast<int>(
                static_cast<long long>(destSampleValue) *
                _mediaInfoBase->audio_format->nSamplesPerSec /
                _mediaAudioDecompressFormat->nSamplesPerSec);
        }

        // p̃obt@쐬EXVB
        // length : 擾\̃TvB
        void UpdateAudioBuffers(int length)
        {
            // f[^ɂTvZo
            int srcLen = CalcSourceAudioSampleValue(length);

            // f[^pobt@쐬EXV
            int srcSize = _mediaInfoBase->audio_format->nBlockAlign * srcLen;
            if (
                _mediaAudioBuffer == nullptr ||
                _mediaAudioBuffer->Length < srcSize)
            {
                _mediaAudioBuffer = gcnew array<Byte>(srcSize);
            }

            // WJpobt@쐬EXV
            int destSize = _mediaAudioDecompressFormat->nBlockAlign * length;
            if (
                _mediaAudioDecompressBuffer == nullptr ||
                _mediaAudioDecompressBuffer->Length < destSize)
            {
                _mediaAudioDecompressBuffer = gcnew array<Byte>(destSize);
            }
        }

      public:
        // vOC̃t@CpX擾B
        property String^ PluginFilePath
        {
            String^ get() { return _filePath; }
        }

        // vOC̃t@C擾B
        property String^ PluginFileName
        {
            String^ get() { return _fileName; }
        }

        // ̓^CvtO擾B
        property InputFlags InputTypes
        {
            InputFlags get() { return _types; }
        }

        // vOC擾B
        property String^ PluginName
        {
            String^ get() { return _pluginName; }
        }

        // t@CtB^擾B
        property String^ FileFilter
        {
            String^ get() { return _fileFilter; }
        }

        // vOC擾B
        property String^ PluginInfo
        {
            String^ get() { return _pluginInfo; }
        }

        // ݒC^tF[XĂ邩ۂ擾B
        property bool HasConfig
        {
            bool get() { return (_pi->func_config != NULL); }
        }

        // YUY2ϊsꍇ̃XP[O^Cv擾܂͐ݒ肷B
        property Yuy2ScaleType Yuy2Scale
        {
            Yuy2ScaleType get() { return _yuy2Scale; }
            void set(Yuy2ScaleType value) { _yuy2Scale = value; }
        };

        // fBAt@CJĂ邩ۂ擾B
        property bool IsMediaFileOpened
        {
            bool get() { return (_mediaInfo != nullptr); }
        }

        // JĂ郁fBAt@C̃tpX擾B
        property String^ MediaFilePath
        {
            String^ get() { return _mediaFilePath; }
        }

        // JĂ郁fBAt@C̓͏擾B
        property IInputInfo^ MediaFileInfo
        {
            IInputInfo^ get() { return _mediaInfo; }
        }

        // fBAt@CJB
        // filePath : fBAt@C̃pXB
        bool OpenMediaFile(String^ filePath)
        {
            // `FbN
            if (filePath == nullptr)
            {
                throw gcnew ArgumentNullException("filePath");
            }

            // ܂
            CloseMediaFile();
            _mediaInfo = nullptr;

            // pX LPSTR ɕϊăt@CJ
            IntPtr pathPtr = Marshal::StringToHGlobalAnsi(filePath);
            _mediaHandle =
                _pi->func_open(static_cast<LPSTR>(pathPtr.ToPointer()));
            Marshal::FreeHGlobal(pathPtr);

            bool result = (_mediaHandle != NULL);
            if (result)
            {
                // tpXݒ
                _mediaFilePath = Path::GetFullPath(filePath);

                // ͏擾
                if (_pi->func_info_get(_mediaHandle, _mediaInfoBase) != FALSE)
                {
                    // ݒ
                    _mediaInfo = gcnew InputInfo(*_mediaInfoBase);

                    // tH[}bgʂ̏
                    InputMediaFlags video = InputMediaFlags::Video;
                    InputMediaFlags audio = InputMediaFlags::Audio;
                    if ((_mediaInfo->InputTypes & video) == video)
                    {
                        result = (result && OpenMediaFileVideo());
                    }
                    if ((_mediaInfo->InputTypes & audio) == audio)
                    {
                        result = (result && OpenMediaFileAudio());
                    }
                }
            }

            // 񂪎擾łĂȂΎs
            result = (result && _mediaInfo != nullptr);
            if (!result)
            {
                CloseMediaFile();
                _mediaFilePath = nullptr;
            }

            return result;
        }

        // JĂ郁fBAt@CB
        bool CloseMediaFile()
        {
            bool result = false;

            if (_mediaHandle != NULL)
            {
                if (_mediaInfo != nullptr)
                {
                    // tH[}bgʂ̉
                    InputMediaFlags video = InputMediaFlags::Video;
                    InputMediaFlags audio = InputMediaFlags::Audio;
                    if ((_mediaInfo->InputTypes & video) == video)
                    {
                        CloseMediaFileVideo();
                    }
                    if ((_mediaInfo->InputTypes & audio) == audio)
                    {
                        CloseMediaFileAudio();
                    }
                }

                // nh̉
                result = (_pi->func_close(_mediaHandle) != FALSE);
                if (result)
                {
                    _mediaInfo = nullptr;
                    _mediaFilePath = nullptr;
                    _mediaHandle = NULL;
                }
            }

            return result;
        }

        // JĂ郁fBAt@C̉摜t[
        // 32rbgARGBgbv_E`Ŏ擾B
        // JĂȂꍇT|[gĂȂꍇ nullptr ԂB
        // f[^͎ɂ̊֐Ăяot@C܂ŕێB
        // frame : t[ʒuB
        array<Byte>^ ReadMediaFileImage(int frame)
        {
            // ԊmF
            InputMediaFlags video = InputMediaFlags::Video;
            if (
                !IsMediaFileOpened ||
                (MediaFileInfo->InputTypes & video) != video)
            {
                return nullptr;
            }

            // `FbN
            if (frame < 0)
            {
                return nullptr;
            }

            // f[^ǂݎ
            pin_ptr<unsigned char> bufPtr = &_mediaImageBuffer[0];
            int size =
                _pi->func_read_video(_mediaHandle, frame, bufPtr);
            bufPtr = nullptr;
            if (size <= 0) { return nullptr; }

            // R[fbNʂ̓WJ
            switch (_mediaImageCodec)
            {
              case VideoCodec::Rgb:
                if (!DecompressRgbImage()) { return nullptr; }
                break;

              case VideoCodec::Yuy2:
                if (!DecompressYuy2Image()) { return nullptr; }
                break;

              case VideoCodec::PixelYc:
                if (!DecompressPixelYcImage()) { return nullptr; }
                break;

              case VideoCodec::Others:
                if (!DecompressOthersImage()) { return nullptr; }
                break;
            }

            return _mediaImageDecompressBuffer;
        }

        // JĂ郁fBAt@C̉Tv
        // 48kHz, 16rbg, XeI`Ŏ擾B
        // JĂȂꍇT|[gĂȂꍇ nullptr ԂB
        // f[^͎ɂ̊֐Ăяot@C܂ŕێB
        // start : 48kHz, 16rbg, XeI`ł̊JnTvʒuB
        // length : 48kHz, 16rbg, XeI`ł̎擾TvB
        array<Byte>^ ReadMediaFileAudio(int start, int length)
        {
            // ԊmF
            InputMediaFlags audio = InputMediaFlags::Audio;
            if (
                !IsMediaFileOpened ||
                (MediaFileInfo->InputTypes & audio) != audio)
            {
                return nullptr;
            }

            // `FbN
            if (start < 0 || length < 0)
            {
                return nullptr;
            }
            if (length == 0)
            {
                return gcnew array<Byte>(1);
            }

            // obt@p
            UpdateAudioBuffers(length);
            pin_ptr<unsigned char> basePtr = &_mediaAudioBuffer[0];
            pin_ptr<unsigned char> destPtr = &_mediaAudioDecompressBuffer[0];

            try
            {
                // f[^ɂʒuƒZo
                int srcStart = CalcSourceAudioSampleValue(start);
                int srcLen = CalcSourceAudioSampleValue(length);

                // f[^ǂݎ
                int size =
                    _pi->func_read_audio(_mediaHandle, srcStart, srcLen, basePtr);
                if (size <= 0) { return nullptr; }

                // ϊpwb_쐬
                ACMSTREAMHEADER convHead;
                ::memset(&convHead, 0, sizeof(ACMSTREAMHEADER));
                convHead.cbStruct = sizeof(ACMSTREAMHEADER);
                convHead.pbSrc = basePtr;
                convHead.cbSrcLength =
                    _mediaInfoBase->audio_format->nBlockAlign * srcLen;
                convHead.pbDst = destPtr;
                convHead.cbDstLength =
                    _mediaAudioDecompressFormat->nBlockAlign * length;

                // ϊ
                MMRESULT res =
                    ::acmStreamPrepareHeader(_mediaAudioHandle, &convHead, 0);
                if (res != 0) { return nullptr; }
                res = ::acmStreamConvert(_mediaAudioHandle, &convHead, 0);
                if (res != 0) { return nullptr; }
                ::acmStreamUnprepareHeader(_mediaAudioHandle, &convHead, 0);
            }
            finally
            {
                destPtr = nullptr;
                basePtr = nullptr;
            }

            return _mediaAudioDecompressBuffer;
        }

        // ݒC^tF[X\B
        // hwnd : eEBhẼnhB nullptr ł悢B
        bool ShowConfig(IntPtr^ handle)
        {
            bool result = false;

            if (HasConfig)
            {
                // nh擾
                HWND hwnd = NULL;
                if (handle != nullptr)
                {
                    hwnd = static_cast<HWND>(handle->ToPointer());
                }

                // Ăяo
                result = (_pi->func_config(hwnd, _hPlugin) != FALSE);
            }

            return result;
        }

        // ̃vOC̕\擾B
        virtual String^ ToString() override
        {
            return PluginName;
        }
    };
}
}}
