using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Drawing;
using System.Drawing.Imaging;
using System.Text;
using System.Windows.Forms;
using System.IO;
using ruche.aviutl.plugins;

namespace ruche.nive2.aviutlPluginPorts
{
    /// <summary>
    /// AviUtlo̓vOCɂo͐ݒs_CAOB
    /// </summary>
    public partial class MediaOutputConfigDialog : Form
    {
        /// <summary>
        /// Ōɐݒ肵YUY2XP[O^CṽL[B
        /// </summary>
        private static readonly string LastYuy2ScaleKeyName =
            typeof(MediaOutputConfigDialog).Name + ":LastYuy2Scale";

        /// <summary>
        /// Ōɐݒ肵YUY2UV̗p@̃L[B
        /// </summary>
        private static readonly string LastYuy2UVKeyName =
            typeof(MediaOutputConfigDialog).Name + ":LastYuy2UV";

        /// <summary>
        /// o̓vOC̃C^tF[XB
        /// </summary>
        private OutputPlugin _plugin = null;

        /// <summary>
        /// o̓vOC̃C^tF[X擾܂͐ݒ肷B
        /// </summary>
        public OutputPlugin Plugin
        {
            get { return _plugin; }
            set
            {
                if (_plugin != value)
                {
                    if (_plugin != null) { _plugin.Dispose(); }
                    _plugin = value;
                }
            }
        }

        public MediaOutputConfigDialog()
        {
            InitializeComponent();
        }

        /// <summary>
        /// o̓vOC[hB
        /// </summary>
        /// <param name="auoFilePath">t@C̃pXB</param>
        private OutputPlugin LoadOutputPlugin(string auoFilePath)
        {
            // `FbN
            if (auoFilePath == null)
            {
                throw new ArgumentNullException("auoFilePath");
            }
            if (!File.Exists(auoFilePath))
            {
                throw new ArgumentException(
                    "w肵o̓vOCt@C݂͑܂B",
                    "auoFilePath");
            }

            OutputPlugin plugin = null;
            try
            {
                // JgfBNg auoFilePath ̏ꏊɂ
                CurrentDirectoryStack.Instance.Push(
                    Path.GetDirectoryName(auoFilePath));

                // o̓vOCC^tF[X쐬
                plugin = new OutputPlugin(auoFilePath);
            }
            finally
            {
                CurrentDirectoryStack.Instance.Pop();
            }

            return plugin;
        }

        /// <summary>
        /// o̓vOCB
        /// </summary>
        private void FreeOutputPlugin()
        {
            if (_plugin != null)
            {
                try
                {
                    // JgfBNgvOC̏ꏊɂ
                    CurrentDirectoryStack.Instance.Push(
                        Path.GetDirectoryName(_plugin.PluginFilePath));

                    _plugin.Dispose();
                    _plugin = null;
                }
                finally
                {
                    CurrentDirectoryStack.Instance.Pop();
                }
            }
        }

        /// <summary>
        /// vOCǂݍ݂ɐƂ̃Rg[XVsB
        /// </summary>
        private void UpdateControlsOnPluginLoaded()
        {
            // Rg[XV
            textPlugin.Text = _plugin.PluginInfo;
            textPlugin.Enabled = true;
            buttonDetail.Enabled = _plugin.HasConfig;
            comboYuy2Scale.Enabled = true;
            comboYuy2UV.Enabled = true;
            buttonOk.Enabled = true;

            // OK{^ANeBuɂ
            this.ActiveControl = buttonOk;
        }

        private void MediaOutputDialog_Load(object sender, EventArgs e)
        {
            // vOCI{^ANeBuɂ
            this.ActiveControl = buttonPlugin;

            if (_plugin != null)
            {
                // Rg[XV
                UpdateControlsOnPluginLoaded();
            }

            // YUY2XP[O^Cv̐ݒ
            comboYuy2Scale.SelectedIndex = (int)Yuy2ScaleType.Compressed;
            if (LibraryConfig.Datas.ContainsKey(LastYuy2ScaleKeyName))
            {
                object value = LibraryConfig.Datas[LastYuy2ScaleKeyName];
                if (value is Yuy2ScaleType)
                {
                    comboYuy2Scale.SelectedIndex = (int)(Yuy2ScaleType)value;
                }
            }

            // YUY2UV̗p@̐ݒ
            comboYuy2UV.SelectedIndex = (int)Yuy2UVType.LeftOrigin;
            if (LibraryConfig.Datas.ContainsKey(LastYuy2UVKeyName))
            {
                object value = LibraryConfig.Datas[LastYuy2UVKeyName];
                if (value is Yuy2UVType)
                {
                    comboYuy2UV.SelectedIndex = (int)(Yuy2UVType)value;
                }
            }
        }

        private void MediaOutputConfigDialog_FormClosing(
            object sender, FormClosingEventArgs e)
        {
            if (_plugin != null)
            {
                // XV
                LibraryConfig.Datas[LastYuy2ScaleKeyName] =
                    (Yuy2ScaleType)comboYuy2Scale.SelectedIndex;
                LibraryConfig.Datas[LastYuy2UVKeyName] =
                    (Yuy2UVType)comboYuy2UV.SelectedIndex;

                // LZȂvOC
                if (DialogResult == DialogResult.Cancel)
                {
                    FreeOutputPlugin();
                }
            }

            // ۑ
            try
            {
                LibraryConfig.Datas.Save();
            }
            catch { }
        }

        private void buttonPlugin_Click(object sender, EventArgs e)
        {
            using (OpenFileDialog dlg = new OpenFileDialog())
            {
                dlg.DefaultExt = "auo";
                dlg.Filter =
                    "AviUtlo̓vOC (*.auo)|*.auo|ׂẴt@C (*.*)|*.*";
                dlg.FilterIndex = 1;
                dlg.Title = "AviUtlo̓vOCI";
                if (_plugin != null)
                {
                    dlg.InitialDirectory =
                        Path.GetDirectoryName(_plugin.PluginFilePath);
                    dlg.FileName = _plugin.PluginFileName;
                }
                else
                {
                    dlg.FileName = "*.auo";
                }

                if (dlg.ShowDialog(this) == DialogResult.OK)
                {
                    try
                    {
                        // vOC̃[h
                        OutputPlugin pi = LoadOutputPlugin(dlg.FileName);
                        FreeOutputPlugin();
                        _plugin = pi;

                        // Rg[XV
                        UpdateControlsOnPluginLoaded();
                    }
                    catch (Exception ex)
                    {
                    #if DEBUG
                        MessageBox.Show(
                            this, ex.Message + "\n\n" + ex.StackTrace,
                            ex.GetType().FullName, MessageBoxButtons.OK);
                    #else
                        MessageBox.Show(
                            this, ex.Message, "G[", MessageBoxButtons.OK);
                    #endif
                    }
                }
            }
        }

        private void buttonDetail_Click(object sender, EventArgs e)
        {
            if (_plugin.HasConfig)
            {
                try
                {
                    // JgfBNgvOC̏ꏊɂ
                    CurrentDirectoryStack.Instance.Push(
                        Path.GetDirectoryName(_plugin.PluginFilePath));

                    // ݒ_CAOĂяo
                    _plugin.ShowConfig(this.Handle);
                }
                finally
                {
                    CurrentDirectoryStack.Instance.Pop();
                }
            }
        }

        private void buttonOk_Click(object sender, EventArgs e)
        {
            // vOCYUY֘Aݒ
            _plugin.Yuy2Scale = (Yuy2ScaleType)comboYuy2Scale.SelectedIndex;
            _plugin.Yuy2UV = (Yuy2UVType)comboYuy2UV.SelectedIndex;

            DialogResult = DialogResult.OK;
        }
    }
}